# include "xr/GCPrivate.h"
# define TICKS_TO_WAIT 2  /* Normal sleep time for daemon */

/*
 * Daemon that invokes collections of partial collections are enabled.
 * If partial collections are enabled, it is the sole initiator of collections.
 * If partial collections are disabled, the mutator invokes the collector
 * by calling GC_full_collection.
 * GC_partial_gc_strategy runs at high priority.  It forks low
 * priority processes to do the background marking.
 */

#define I_HOLD_ML(ml) (((XR_ML)(ml))->ml_holder == XR_currThread)

struct XR_CVRep neededCV = { 0 };

bool GC_stop_world_collection_needed = FALSE;
					/* Full stop-the-world collection  */
					/* requested asap.		   */
bool GC_full_collection_needed = FALSE; /* Full collection has been	*/
					/* requested.			*/
					
bool GC_Need_incremental_gcP()
{
    register int c_in_use;
    register int adj_words_allocd;
    
    if (GC_mode & GC_INCREMENTAL) {
	/* Test whether allocation warrants a collection */
	    /* We count GC_mem_free almost as though it had never 	*/
	    /* been allocated.  We don't quite want to do this,  	*/
	    /* since a nonleaking client still needs to have storage	*/
	    /* coalesced occasionally.  This is a bit time-critical;	*/
	    /* hence the hack to avoid multiplication and division.	*/
	    adj_words_allocd = GC_words_allocd - GC_mem_freed
	    				       + (GC_mem_freed >> 3);
	    c_in_use = GC_composite_at_full_gc;
	    if (c_in_use < MIN_COMPOSITE_IN_USE) {
	      c_in_use = MIN_COMPOSITE_IN_USE;
	    }   
	    if (GC_partial_gc_allocs) {
	      return (adj_words_allocd > GC_partial_gc_allocs);
	    } else {
	      return (adj_words_allocd * GC_free_mem_ratio > c_in_use);
	    }
    } else {
	/* not incremental, so we'll only collect on demand */
	return(FALSE);
    }
}

/* Delay caller sufficiently to allow collector to catch up.  Caller is */
/* assumed to have allocated nbytes, and may be penalized 		*/
/* proportionately.							*/
/* Caller should avoid holding important locks.				*/
/* Caller should not hold allocation lock.				*/
void GC_delay_alloc(nbytes)
long nbytes;
{
    static int counter = 0;
    register int wait_ticks;
#   define ONE_TICK_PENALTY 32768  /* This many bytes per 1 tick penalty   */
#   define SMALL_OBJ_INTERVAL 8	   /* 1 in every SMALL_OBJ_INTERVAL small  */
				   /* object allocations will be penalized */
				   /* one tick.				   */
#   define MAX_PENALTY 2 	   /* Maximum introduced delay. 	   */
    
    if (GC_Need_incremental_gcP() && !GC_dont_gc) {
        if (nbytes > ONE_TICK_PENALTY) {
            XR_MonitorEntry(&GC_allocate_ml);
            XR_Notify(&neededCV);
            wait_ticks = nbytes/ONE_TICK_PENALTY;
            if (wait_ticks > MAX_PENALTY) {
                wait_ticks = MAX_PENALTY;
            }
            GC_wait_for_gc(wait_ticks);
            XR_MonitorExit(&GC_allocate_ml);
        } else {
            /* delay every once in a while */
            counter ++;
            if (counter % SMALL_OBJ_INTERVAL == 0) {
                XR_MonitorEntry(&GC_allocate_ml);
                XR_Notify(&neededCV);
                GC_wait_for_gc(1);
                XR_MonitorExit(&GC_allocate_ml);
            }
        }
    }
}

void
GC_wait_until_gc_needed()
{
  while(1) {
    /* Wait a while to give mutator a chance */
    XR_WaitCV(&neededCV, NIL);

    /* collections allowed at all? */
    if (!GC_dont_gc) {
      if (GC_demand_collection) {
	  GC_demand_collection = FALSE;
	  return;
      } else {
	  if (GC_Need_incremental_gcP()) {
	      return;
	  }
      }
    }
    /* wait for the next one. */
  }
}

bool
GC_Need_full_gcP()
{
  register int adj_words_allocd;
  
  if (!(GC_mode & GC_INCREMENTAL)) return(TRUE);
  if (GC_full_collection_needed) return(TRUE);
  if (GC_full_gc_allocs != 0) {
      /* See explanation in Need_incremental_gcP */
      adj_words_allocd = GC_words_allocd - GC_mem_freed
	    				 + GC_mem_freed >> 3;
      return(adj_words_allocd > GC_full_gc_allocs);
  } else {
      return(GC_composite_in_use + GC_atomic_in_use
             > GC_words_at_full_gc
               + GC_composite_at_full_gc/GC_free_mem_ratio);
  }
}

# define TOO_MANY_FAULTS 1000  /* Threshold for switching to careful */
                               /* marking.			     */

bool GC_should_do_second_mark(initial_words_allocd) 
unsigned long initial_words_allocd;
{
#   define SECOND_MARK_THRESHOLD 10000
	/* If more than this many words are allocated during first par.	*/
	/* marking pass, we run a second one.				*/
    return (GC_words_allocd - initial_words_allocd > SECOND_MARK_THRESHOLD);
}

bool GC_should_mark_carefully()
{
  return (GC_max_markfaults > TOO_MANY_FAULTS);
}

bool GC_should_pre_collect()
{
  /* should do better than this.  We can also look to see if the partial
     collection seems likely to free up much storage and look to see
     if the full collection is likely to finish before storage runs
     out. */
  return (GC_mode & GC_PARALLEL);
}

/* Make sure that GC_mode agrees with what was requested */
void GC_update_mode()
{
    XR_MonitorEntry( &GC_allocate_ml );
    if (GC_requested_mode != GC_mode) {
        if ((GC_mode & DIRTY_BITS_REQUIRED)
            && !(GC_requested_mode & DIRTY_BITS_REQUIRED)) {
            XR_EndVD();
        }
        if (!(GC_mode & DIRTY_BITS_REQUIRED)
            && (GC_requested_mode & DIRTY_BITS_REQUIRED)) {
            XR_StartVD();
        }
        GC_mode = GC_requested_mode;
    }
    XR_MonitorExit( &GC_allocate_ml );
}

void GC_daemon()
{
  XR_SetPriority(XR_PRI_SYS_FOREGROUND);
  GC_lock_to_fixed_vp();
  	/* We need to run on the same processor as marker threads.  	*/
  	/* Otherwise we can't do directed yields to them.		*/
  XR_InitializeCondition(&neededCV, TICKS_TO_WAIT);
# ifdef DIRTY_BITS
    XR_GCSetMode(GC_PARALLEL | GC_INCREMENTAL);
# endif

  GC_wait_until_gc_needed();

  while(1) {
    GC_update_mode();
    GC_demand_collection = FALSE;
    GC_full_collection(GC_stop_world_collection_needed);
    GC_stop_world_collection_needed = FALSE;
    GC_full_collection_needed = FALSE;

    GC_words_at_full_gc = GC_my_atomic_in_use + GC_my_composite_in_use;
    GC_composite_at_full_gc = GC_my_composite_in_use;
    GC_wait_until_gc_needed();

    while(!GC_Need_full_gcP()) {
      GC_demand_collection = FALSE;
      GC_partial_collection();
      GC_wait_until_gc_needed();
    }
  }
  /* We don't ever get here. */
}


# define MAXHEAPWASTE 2

/* Request a collection */
void GC_demand_and_dont_wait()
{
    GC_demand_collection = TRUE;
    GC_start_daemon();
    XR_Notify(&neededCV);
}

/* assumes that the allocate lock is held       */
/* This starts a full stop-the-world collection */
void GC_demand_full_and_wait()
{
    GC_stop_world_collection_needed = TRUE;
    GC_full_collection_needed = TRUE;
    GC_demand_and_dont_wait();
    GC_wait_for_gc(XR_WAIT_FOREVER);
}

bool GC_should_collect_on_heap_full()
{
    if (GC_dont_gc) return(FALSE);
    if (GC_mode & (GC_INCREMENTAL|GC_PARALLEL)) {
        return(MAXHEAPWASTE*(GC_composite_in_use + GC_atomic_in_use)
               < BYTES_TO_WORDS(GC_heapsize) );
    } else {
        return(TRUE);
    }
}

/* Expand heap if there was a recent collection, otherwise collect */
/* We hold GC_allocate_ml					   */
static void GC_expand_or_collect(min_words)
register long min_words;
{
    register long min_blocks =
          divHBLKSZ(WORDS_TO_BYTES(min_words) + 2 * HBLKSIZE - 1);
          
    if (GC_words_allocd * GC_free_mem_ratio > GC_composite_in_use) {
        GC_demand_full_and_wait();
    } else {
        int blocks_to_get = divHBLKSZ(
        			WORDS_TO_BYTES(
        			    GC_composite_in_use
        			    - GC_words_allocd * GC_free_mem_ratio)
        			+ 2 * HBLKSIZE);
        if ( blocks_to_get > MAX_HINCR ) blocks_to_get = MAX_HINCR;
        if ( blocks_to_get < MIN_HINCR ) blocks_to_get = MIN_HINCR;
        if ( min_blocks  > blocks_to_get ) blocks_to_get = min_blocks;
        GC_expand_hp(blocks_to_get, FALSE);
    }
}

/* assumes that the allocate lock is held */
void GC_heap_full()
{
  /* well, I don't like this much.
     First try to free some blocks.  If THAT fails, demand a collection */
  GC_reclaim_empty_pages();
  if (!GC_sufficient_hb(1)) {
    if (GC_should_collect_on_heap_full()) {
      GC_expand_or_collect(1);
    }
  }
}

/* assumes that the allocate lock is held */
void GC_reclaim_hblks(lw)
long lw;
{
  /* well, I don't like this much either.
     First try to free some blocks.  If THAT fails, demand a collection */
  GC_reclaim_empty_pages();
  if (!GC_sufficient_hb(lw)) {
    if (GC_should_collect_on_heap_full()) {

      GC_expand_or_collect(abs(lw));

      if (!(GC_sufficient_hb(lw))) {
	GC_reclaim_empty_pages();
	/* should we have something here demanding a full collection? */
      }
    }
  }
}

void XR_GCollect()
{
  XR_MonitorEntry(&GC_allocate_ml);
  GC_demand_full_and_wait();
  XR_MonitorExit(&GC_allocate_ml);
}

